@@ -31,6 +31,18 @@
  *
  *  ATTENTION!   ATTENTION!   ATTENTION!   ATTENTION!   ATTENTION! */
 
+/* 4-01-2002
+ * Initialize thread locks for fake files in vsnprintf and vdprintf.
+ *    reported by Erik Andersen (andersen@codepoet.com)
+ * Fix an arg promotion handling bug in _do_one_spec for %c. 
+ *    reported by Ilguiz Latypov <ilatypov@superbt.com>
+ *
+ * 5-10-2002
+ * Remove __isdigit and use new ctype.h version.
+ * Add conditional setting of QUAL_CHARS for size_t and ptrdiff_t.
+ */
+
+
 #define _ISOC99_SOURCE			/* for ULLONG primarily... */
 #define _GNU_SOURCE
 #define _STDIO_UTILITY			/* We're using _uintmaxtostr. */
@@ -54,7 +66,12 @@
 
 /**********************************************************************/
 
+/* These provide some control over printf's feature set */
 #define __STDIO_PRINTF_FLOAT
+#define __STDIO_PRINTF_M_SUPPORT
+
+
+/**********************************************************************/
 
 #if defined(__UCLIBC__) && !defined(__UCLIBC_HAS_FLOATS__)
 #undef __STDIO_PRINTF_FLOAT
@@ -118,7 +135,9 @@ enum {
 	CONV_x, CONV_X,	CONV_o,	CONV_u,	CONV_d,	CONV_i,
 	CONV_f, CONV_F, CONV_e, CONV_E, CONV_g, CONV_G, CONV_a, CONV_A,
 	CONV_C, CONV_S, CONV_c, CONV_s,
+#ifdef __STDIO_PRINTF_M_SUPPORT
 	CONV_m,
+#endif
 	CONV_custom0				/* must be last */
 };
 
@@ -158,12 +177,6 @@ enum {
  */
 
 /* TODO -- Fix the table below to take into account stdint.h. */
-#if PTRDIFF_MAX != INT_MAX
-#error fix QUAL_CHARS ptrdiff_t entry 't'!
-#endif
-#if SIZE_MAX != UINT_MAX
-#error fix QUAL_CHARS size_t entries 'z', 'Z'!
-#endif
 #ifndef LLONG_MAX
 #error fix QUAL_CHARS for no long long!  Affects 'L', 'j', 'q', 'll'.
 #else
@@ -172,11 +185,38 @@ enum {
 #endif
 #endif
 
+#ifdef PDS
+#error PDS already defined!
+#endif
+#ifdef SS
+#error SS already defined!
+#endif
+
+#if PTRDIFF_MAX == INT_MAX
+#define PDS		0
+#elif PTRDIFF_MAX == LONG_MAX
+#define PDS		4
+#elif defined(LLONG_MAX) && (PTRDIFF_MAX == LLONG_MAX)
+#define PDS		8
+#else
+#error fix QUAL_CHARS ptrdiff_t entry 't'!
+#endif
+
+#if SIZE_MAX == UINT_MAX
+#define SS		0
+#elif SIZE_MAX == ULONG_MAX
+#define SS		4
+#elif defined(LLONG_MAX) && (SIZE_MAX == ULLONG_MAX)
+#define SS		8
+#else
+#error fix QUAL_CHARS size_t entries 'z', 'Z'!
+#endif
+
 #define QUAL_CHARS		{ \
 	/* j:(u)intmax_t z:(s)size_t  t:ptrdiff_t  \0:int */ \
 	/* q:long_long  Z:(s)size_t */ \
 	'h',   'l',  'L',  'j',  'z',  't',  'q', 'Z',  0, \
-	 2,     4,    8,    8,    0,    0,    8,   0,   0, /* TODO -- fix!!! */\
+	 2,     4,    8,    8,   SS,  PDS,    8,  SS,   0, /* TODO -- fix!!! */\
      1,     8 \
 }
 
@@ -247,9 +287,6 @@ typedef struct {
 /* TODO: fix printf to return 0 and set errno if format error.  Standard says
    only returns -1 if sets error indicator for the stream. */
 
-/* TODO -- __isdigit() macro */
-#define __isdigit(c) (((unsigned int)(c - '0')) < 10)
-
 #ifdef __STDIO_PRINTF_FLOAT
 extern size_t _dtostr(FILE * fp, long double x, struct printf_info *info);
 #endif
@@ -878,11 +915,13 @@ extern int _ppfs_parsespec(ppfs_t *ppfs)
 	if (!*p) {
 #ifdef __STDIO_GLIBC_CUSTOM_PRINTF
 		/* TODO -- gnu %m support build option. */
+#ifdef __STDIO_PRINTF_M_SUPPORT
 		if (*fmt == 'm') {
 			conv_num = CONV_m;
 			ppfs->num_data_args = 0;
 			goto DONE;
 		}
+#endif
 
 		/* Handle custom arg -- WARNING -- overwrites p!!! */
 		conv_num = CONV_custom0;
@@ -1064,7 +1103,7 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 	}
 #endif
 	{
-		register char *s;
+		register char *s;		/* TODO: Should s be unsigned char * ? */
 
 		if (ppfs->conv_num == CONV_n) {
 			_store_inttype(*(void **)*argptr,
@@ -1162,13 +1201,15 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 				}
 			} else {			/* char */
 				s = (char *) buf;
-				*s = *((const char *) *argptr);
+				*s = (unsigned char)(*((const int *) *argptr));
 				s[1] = 0;
 				slen = 1;
 			}
+#ifdef __STDIO_PRINTF_M_SUPPORT
 		} else if (ppfs->conv_num == CONV_m) {
-			s = strerror(errno);
+			s = _stdio_strerror_r(errno, buf, sizeof(buf));
 			goto SET_STRING_LEN;
+#endif
 		} else {
 			assert(ppfs->conv_num == CONV_custom0);
 
@@ -1253,6 +1294,10 @@ int vsnprintf(char *__restrict buf, size_t size,
 	f.filedes = -2;				/* for debugging */
 	f.modeflags = (__FLAG_NARROW|__FLAG_WRITEONLY|__FLAG_WRITING);
 
+#ifdef __STDIO_THREADSAFE
+	__stdio_init_mutex(&f.lock);
+#endif
+
 	rv = vfprintf(&f, format, arg);
 	if (size) {
 		if (f.bufwpos == f.bufend) {
@@ -1323,6 +1368,10 @@ int vsnprintf(char *__restrict buf, size_t size,
 	f.filedes = -1;				/* For debugging. */
 	f.modeflags = (__FLAG_NARROW|__FLAG_WRITEONLY|__FLAG_WRITING);
 
+#ifdef __STDIO_THREADSAFE
+	__stdio_init_mutex(&f.lock);
+#endif
+
 	rv = vfprintf(&f, format, arg);
 
 	return rv;
@@ -1361,6 +1410,10 @@ int vdprintf(int filedes, const char * __restrict format, va_list arg)
 	f.filedes = filedes;
 	f.modeflags = (__FLAG_NARROW|__FLAG_WRITEONLY|__FLAG_WRITING);
 
+#ifdef __STDIO_THREADSAFE
+	__stdio_init_mutex(&f.lock);
+#endif
+
 	rv = vfprintf(&f, format, arg);
 
 	return fflush(&f) ? -1 : rv;
@@ -2038,7 +2091,7 @@ extern uintmax_t _load_inttype(int desttype, const void *src, int uflag)
 #if LONG_MAX != INT_MAX
 		if (desttype & (PA_FLAG_LONG|PA_FLAG_LONG_LONG)) {
 #ifdef LLONG_MAX
-			if (destsize == PA_FLAG_LONG_LONG) {
+			if (desttype == PA_FLAG_LONG_LONG) {
 				return *((long long int *) src);
 			}
 #endif
